import aiohttp
import asyncio
from functools import partial
from yarl import URL
from nougat.protocol import HttpProtocol
try:
    import uvloop
except:
    uvloop = asyncio


__all__ = ['TestClient']


HOST = '127.0.0.1'
PORT = 40404


class Proxy:
    ret = None


class TestClient:
    app = None
    loop = None
    server = None

    def __init__(self, app=None):
        self.app = app
        self.loop = uvloop.new_event_loop()
        self.server = None

    def __request(self, method, url, *args, **kwargs):

        ret = Proxy()

        async def test_middleware(ctx, next):
            await next()
            ctx.app.ctx = ctx

        def response_builder(_ret):
            def _response(future):
                _ret.ret = future.result()
                asyncio.get_event_loop().stop()
            return _response

        response_function = response_builder(ret)

        async def __local_request(method, url, future, *args, **kwargs):
            url = 'http://{host}:{port}{uri}'.format(host=HOST, port=PORT, uri=url)
            async with aiohttp.ClientSession(loop=self.loop) as session:
                async with getattr(session, method)(url, *args, **kwargs) as response:
                    response.text = await response.text()
                    future.set_result(response)
                    asyncio.get_event_loop().stop()

        self.app.chain.append(test_middleware)
        asyncio.set_event_loop(self.loop)

        server_coroutine = self.loop.create_server(partial(HttpProtocol, loop=self.loop, app=self.app), HOST, PORT)
        server_loop = self.loop.run_until_complete(server_coroutine)

        # set request
        request_future = asyncio.Future()
        asyncio.ensure_future(__local_request(method, url, request_future, *args, **kwargs), loop=self.loop)
        request_future.add_done_callback(response_function)
        self.loop.run_until_complete(request_future)

        self.loop.run_forever()

        server_loop.close()
        if not hasattr(self.app, "ctx"):
            self.app.ctx = None

        return ret.ret, self.app.ctx

    def head(self, url, *args, **kwargs):
        return self.__request('head', url, *args, **kwargs)

    def get(self, url, *args, **kwargs):
        return self.__request('get', url, *args, **kwargs)

    def post(self, url, *args, **kwargs):
        return self.__request('post', url, *args, **kwargs)

    def put(self, url, *args, **kwargs):
        return self.__request('put', url, *args, **kwargs)

    def delete(self, url, *args, **kwargs):
        return self.__request('delete', url, *args, **kwargs)

    def options(self, url, *args, **kwargs):
        return self.__request('options', url, *args, **kwargs)

    def patch(self, url, *args, **kwargs):
        return self.__request('patch', url, *args, **kwargs)

    def url(self, path=""):
        """
        generate the abstract url for test case
        """

        return URL('http://{}:{}{}'.format(HOST, PORT, path))



import asyncio
import logging
import signal
import sys
import inspect
from functools import partial
from nougat.router import Router, Param
from nougat.protocol import HttpProtocol
from nougat.test_client import TestClient
from nougat.section import Section
from nougat.exceptions import *
from nougat.utils import is_middleware
from nougat.config import Config

try:
    import uvloop
except:
    uvloop = asyncio

__all__ = ['Nougat']


class Nougat(object):

    def __init__(self, name=None):

        self.name = name
        self.__test_client = None
        self.router = Router(self.name)

        self.config = Config()
        self.chain = []

        self.sections = {}

    def use(self, middleware_or_section_name):
        """
        register middleware or section 
        """
        if isinstance(middleware_or_section_name, Section):
            # register section
            section = middleware_or_section_name
            if section.name in self.sections:
                raise NougatRuntimeError("it seems that this section's name had been used")
            logging.debug("register section {}".format(section.name))
            self.sections[section.name] = section

            self.router.union(section.router)
        elif inspect.isfunction(middleware_or_section_name):
            # register middleware
            is_middleware(middleware_or_section_name)
            middleware = middleware_or_section_name
            self.chain.insert(0, middleware)
        else:
            raise NougatRuntimeError("nougat only can use section instance or middleware function")

    async def handler(self, context, handler_future):

        try:
            # find route
            route = self.router.get(context)
            if not route:
                raise HttpException(None, 404)

            section = self.sections.get(route.section_name)
            context.generate_params(route)

            handler = partial(section.handler, context=context, route=route)

            for middleware in self.chain:
                handler = partial(middleware, ctx=context, next=handler)

            await handler()

        except HttpException as e:

            context.status = e.status
            context.res = e.body

        handler_future.set_result(context)

    @property
    def test(self):
        if not self.__test_client:
            return TestClient(self)
        return self.__test_client

    def run(self, host="127.0.0.1", port=8000, debug=False, loop=None):
        # Create Event Loop
        loop = loop or uvloop.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.set_debug(debug)

        if debug:
            logging.basicConfig(level=logging.DEBUG)

        def ask_exit():
            loop.stop()

        if sys.platform != 'win32':
            for signame in ('SIGINT', 'SIGTERM'):
                loop.add_signal_handler(getattr(signal, signame), ask_exit)

        server_coroutine = loop.create_server(partial(HttpProtocol, loop=loop, app=self), host, port)
        server_loop = loop.run_until_complete(server_coroutine)
        try:
            print("server http://{}:{} is running, press Ctrl+C to interrupt.".format(host, port))
            loop.run_forever()
        finally:
            server_coroutine.close()
            loop.close()

    def stop(self):
        asyncio.get_event_loop().stop()



import re
from nougat.exceptions import UnknownRouterException, RouteReDefineException, \
    ParamMissingException, ParamRedefineException


__all__ = ['Router', 'Param', 'Route', 'DynamicRoute', 'StaticRoute']


METHODS = ['HEAD', 'GET', 'POST', 'PUT', 'DELETE', 'HEAD', 'PATCH', 'OPTIONS']


class Router:
    """
    路由系统
    """

    def __init__(self, app_name):

        self.__app_name = app_name

        self.fixed_routes = {}  # 静态路由
        self.dynamic_routes = {}  # 动态路由

        self.__init_route()
        self.__dynamic_pattern = re.compile(r"(<([a-zA-Z_]+)>)+")

    def __init_route(self):
        """
        初始化 静态路由、动态路由
        """
        for method in METHODS:
            self.fixed_routes[method] = {}
            self.dynamic_routes[method] = []

    def add(self, rule, handler, section_name, method, params):
        """
        添加路由
        :param rule: 路由规则
        :param handler: 处理方法
        :param section_name: the name of section
        :param method: 请求方法
        :param params: Param 列表
        """
        is_dynamic, match_pattern, url_params = self.__is_dynamic(rule)

        if is_dynamic:
            this_route = DynamicRoute(rule, handler, section_name, match_pattern, url_params)

            # param missing check
            params_key = [param.name for param in params]
            x = [param for param in this_route.url_params if param not in params_key]
            if x:
                raise ParamMissingException(rule, x[0])

            self.dynamic_routes[method].append(this_route)

        else:

            this_route = StaticRoute(rule, handler, section_name)

            # raise rule redefine
            if rule in self.fixed_routes[method]:
                raise RouteReDefineException(method, rule)

            self.fixed_routes[method][rule] = this_route

        for param in params:
            this_route.add_param(**param)

        return this_route

    def __is_dynamic(self, rule):
        """
        判断路由是否动态
        :param rule: 路由规则
        :return: boolean, formated_rule
        """
        match = self.__dynamic_pattern.findall(rule)
        if match:
            ret = rule
            params = []
            for param in match:
                ret = ret.replace(param[0], "(?P<{}>[^/]+)".format(param[1]))
                if param[1] in params:
                    raise ParamRedefineException(rule, param[1])

                params.append(param[1])
            return True, ret, params

        return False, rule, None

    def get(self, context):
        """
        寻找与 url 匹配的路由
        :param context: request context
        :return: Route 类
        """

        method = context.method
        url = context.url.path

        # try finding route in static map
        route = self.fixed_routes[method].get(url, None)

        if route is not None:
            return route

        else:
            # matching from dynamic routes

            for one_route in self.dynamic_routes[method]:
                if one_route.match(url):
                    return one_route

        return route

    def union(self, router):
        if not isinstance(router, Router):
            raise UnknownRouterException()

        for method in METHODS:

            # try get inner dict
            inter = [x for x in self.fixed_routes[method] if x in router.fixed_routes[method]]
            if inter:
                raise RouteReDefineException(method, inter[0])
            self.fixed_routes[method].update(router.fixed_routes[method])

            self.dynamic_routes[method].extend(router.dynamic_routes[method])


class Param:

    def __init__(self, name, type, location=None, optional=False, default=None, action=None, append=False, description=None):
        self.name = name  # name
        self.type = type  # type or [type, type]
        self.location = location  # cookies, query, form, headers
        self.optional = optional  # true, false
        self.default = default  # if optional is true
        self.action = action  # rename
        self.append = append  # list or not
        self.description = description  # description

        # location iterable
        if not isinstance(self.location, list):
            self.location = [self.location]


class Route:
    """
    基本路由规则
    """
    rule = None
    handler = None
    params = {}
    section_name = ''

    def add_param(self, name, type, **kwargs):
        """
        向规则中添加参数
        """
        if name in self.params:
            raise ParamRedefineException(self.rule, name)
        self.params[name] = Param(name, type, **kwargs)

    def url(self, **kwargs):
        pass


class DynamicRoute(Route):
    """
    动态路由规则
    """

    def __init__(self, rule, handler, section_name, pattern, url_params):
        self.rule = rule
        self.handler = handler
        self.section_name = section_name
        self.pattern = re.compile(pattern)
        self.params = {}
        self.url_params = url_params
        self.url_params_dict = {}

    def match(self, url):
        match = self.pattern.match(url)
        if match:
            if match.pos == 0 and match.endpos == len(url):
                self.url_params_dict = match.groupdict()
                return match

        return None

    def url(self, **kwargs):
        url_ret = self.rule
        for key, _ in self.url_params_dict.items():
            value = kwargs.get(key, None)
            if not value:
                _param = self.params.get(key, None)
                if _param:
                    value = _param.default


            if not value:
                raise Exception()  # TODO param miss exception

            url_ret = url_ret.replace("<{}>".format(key), value)
            kwargs.pop(key)

        if kwargs:
            url_ret = "{}?{}".format(url_ret, "&".join(["{}={}".format(key, value) for key, value in kwargs.items()]))

        return url_ret


class StaticRoute(Route):
    """
    静态路由规则
    """
    def __init__(self, rule, handler, section_name):
        self.rule = rule
        self.handler = handler
        self.section_name = section_name
        self.params = {}

    def url(self, **kwargs):
        url_ret = self.rule
        if kwargs:
            url_ret = "{}?{}".format(url_ret, "&".join(["{}={}".format(key, value) for key, value in kwargs.items()]))
        return url_ret



import asyncio
import httptools
from nougat.context import Context

__all__ = ['HttpProtocol']


class HttpProtocol(asyncio.Protocol):

    def __init__(self, app, loop):
        """
        :param app: Nougat Instance
        :param loop: uvloop or asyncio loop
        """
        self.app = app
        self.loop = loop

        self.path = None
        self.headers = None

        self.context = None

        self.transport = None
        self.parser = None

        self.__body = []

    def connection_made(self, transport):
        """
        asyncio.Protocol override 函数，当 Protocol 连接时调用
        """
        self.transport = transport

    def connection_lost(self, exc):
        """
        asyncio.Protocol override 函数，当 Protocol 失去连接时调用
        """
        self.context = None
        self.parser = None
        self.__body = []

    def data_received(self, data):
        """
        asyncio.Protocol override 函数，当 数据传入时调用

        """

        if self.parser is None:
            self.headers = []
            self.parser = httptools.HttpRequestParser(self)

        # 传递信息至 HttpRequestParser 解析
        self.parser.feed_data(data)

    def on_url(self, url):
        """
        HttpRequestParser 解析函数
        """
        self.path = url

    def on_header(self, key, value):
        """
        HttpRequestParser 解析函数
        """
        self.headers.append((key.decode(), value.decode()))

    def on_headers_complete(self):
        """
        HttpRequestParser 解析函数
        """
        # 构建 Context 类
        self.context = Context(
            app=self.app,
            path=self.path,
            headers=self.headers,
            ip=self.transport.get_extra_info('peername')[0],
            version=self.parser.get_http_version(),
            method=self.parser.get_method().decode('utf-8'),
        )

    def on_body(self, body):
        self.__body.append(body)

    def on_message_complete(self):

        self.context.init_body(b''.join(self.__body))

        # calling handler function
        handler_future = asyncio.Future()
        asyncio.ensure_future(self.app.handler(self.context, handler_future))
        handler_future.add_done_callback(self.on_response)

    def on_response(self, handler_future):

        response = handler_future.result()
        self.transport.write(response.output)

        if not self.parser.should_keep_alive():
            self.transport.close()
        self.parser = None
        self.context = None
